// test.c
#define F_CPU 16000000
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

#include <stdbool.h>
#include <stdio.h>
#include <string.h>

#include "pff.h"

#define MUSIC "music"
#define WRITE_BUF_SIZE	256

FATFS fs;	// sistemul de fisiere

char filename[32] = { 0 };
unsigned char buffer1[256]; /* wave output fifo */
unsigned char buffer2[256]; /* wave output fifo */
char file_no = 1;
int sample = 0;
int recording = 1;
int playing = 1;

WORD rb;
volatile uint8_t read1 = 1;
volatile uint16_t counter=0;
volatile uint8_t sync = 0;
volatile uint8_t write1 = 1;
volatile uint8_t rec_sample;
volatile uint32_t adc_val;
volatile uint32_t converted;
volatile uint32_t sent;

/* debugging cu ledul: */
void led_red()
{
	int i;
	for (i = 0; i < 4; i++) {
		PORTD ^= (1 << PD2);
		_delay_ms(500);
	}
	PORTD &= ~(1 << PD2);
}

void led_green()
{
	int i;
	for (i = 0; i < 4; i++) {
		PORTD ^= (1 << PD1);
		_delay_ms(500);
	}
	PORTD &= ~(1 << PD1);
}

void led_blue()
{
	int i;
	for (i = 0; i < 4; i++) {
		PORTD ^= (1 << PD0);
		_delay_ms(500);
	}
	PORTD &= ~(1 << PD0);

}

void show_number(int n)
{
	int i;
	for (i = 0; i < 2 * n; i++) {
		PORTD ^= (1 << PD0);
		PORTD ^= (1 << PD1);
		_delay_ms(500);	
	}
}

void IO_init(void)
{
	/* init LED-uri */
    DDRD |= (1 << PD0);
    DDRD |= (1 << PD1);
    DDRD |= (1 << PD2);
    DDRD |= (1 << PD3);
    DDRC |= (1 << PC0);
    DDRC |= (1 << PC1);
    DDRC |= (1 << PC2);
    DDRC |= (1 << PC3);


	/* stinge LED-uri */
    PORTD &= ~(1 << PD0);
    PORTD &= ~(1 << PD1);
    PORTD &= ~(1 << PD2);
    PORTD &= ~(1 << PD3);
    PORTC &= ~(1 << PC0);
    PORTC &= ~(1 << PC2);

	/* aprinde LED-uri pentru ochi */
    PORTC |= (1 << PC1);
    PORTC |= (1 << PC3);

	/* init senzor sunet: PC6 pin de input, fara rezistenta pull-up */
    DDRC &= ~(1 << PC6);
    PORTC &= ~(1 << PC6);

	/* init difuzor */
    DDRD |= (1 << PD4);

	/* init microfon: PC5-7 pini de input, fara rezistenta pull-up */
    DDRA &= ~(1 << PA5);
    DDRA &= ~(1 << PA6);
    DDRA &= ~(1 << PA7);

    PORTA &= ~(1 << PA5);
    PORTA &= ~(1 << PA6);
    PORTA &= ~(1 << PA7);

}

ISR(ADC_vect){
	// take sample value
	adc_val = ADC;
	adc_val -= 0x255;

	// normalize value
	
	if (write1){
		buffer1[counter++] = adc_val;
		if (counter == WRITE_BUF_SIZE){
			counter = 0;
			converted ++;
			write1 = 0;
		}
	}
	else {
		buffer2[counter++] = adc_val;
		if (counter == WRITE_BUF_SIZE){
			counter = 0;
			converted ++;
			write1 = 1;
		}
	}
}

void ADC_init(void)
{
	ADMUX = (1 << REFS0) | (1 << MUX1) | (1 << MUX2);
	ADCSRA = (1 << ADEN) | (1 << ADSC) | (1 << ADATE) | (1 << ADIE) | (1 << ADPS2);
	ADCSRB |= (1 << ADTS1) | (1 << ADTS0);
}


/*---------------------------------------------------------------------------*/
/* Player audio                                                              */
/*---------------------------------------------------------------------------*/

/*
 * Four-Character Code - folosit pentru a indentifica formate de date
 */
#define FCC(c1, c2, c3, c4) \
	(((DWORD)(c4) << 24) + \
	 ((DWORD)(c3) << 16) + \
	 (( WORD)(c2) <<  8) + \
	 (( BYTE)(c1) <<  0))


	 uint8_t	buf[2][256];	// wave output buffers (double buffering)
const	 uint16_t	buf_size = 256;	// front and back buffer sizes
volatile uint8_t	buf_front = 0;	// front buffer index (current buffer used)
volatile uint8_t	buf_pos = 0;	// current buffer position
volatile uint8_t	buf_sync = 0;
volatile bool flag = false;
volatile int ctr = 0;

#define BUF_FRONT	(buf[buf_front])
#define BUF_BACK	(buf[1 - buf_front])


ISR(TIMER0_COMPA_vect)
{
	OCR1B = BUF_FRONT[buf_pos++];

	// swap buffers when end is reached (end is 256 <=> overflow to 0)
	if(buf_pos == 0)
		buf_front = 1 - buf_front;

}

ISR(TIMER2_OVF_vect){

// DEBUGGING:	led_blue();

    if (ctr > 1){
        ctr--;
    } else if (ctr == 1) {
        flag = false;
    	PORTD &= ~(1 << PD1);
    }

}

ISR(PCINT2_vect)
{

// DEBUGGING:	led_blue();

    flag = true;
    ctr = 120;
}

void timer0_start(void)
{
	// interrupt on compare A
	TIMSK0 |= (1 << OCIE0A);
	// CTC, top OCRA
	TCCR0B |= (0 << WGM02);
	TCCR0A |= (1 << WGM01) | (0 << WGM00);
	// prescaler 8
	TCCR0B |= (2 << CS00);
}

void timer0_stop(void)
{
	TCCR0B = 0;
	TCCR0A = 0;
	TIMSK0 = 0;
	OCR0A = 0;
	TCNT0 = 0;
}

void timer1_start(void)
{
	// 8-bit FastPWM
	TCCR1B |= (1 << WGM12);
	TCCR1A |= (1 << WGM10);
	// channel B inverted
	TCCR1A |= (1 << COM1B0) | (1 << COM1B1);
	// prescaler 1
	TCCR1B |= (1 << CS10);
}

void timer1_stop(void)
{
	TCCR1B = 0;
	TCCR1A = 0;
	OCR1B = 0;
	TCNT1 = 0;
}

// folosit pentru detectia intensitatii sunetului
void timer2_start(void)
{
	// interrupt on compare A
	TIMSK2 |= (1 << TOIE2);
	// prescaler 1024
	TCCR2B |= (7 << CS20);
}

void timer2_stop(void)
{
	TCCR2B = 0;
	TCCR2A = 0;
	TIMSK2 = 0;
	OCR2A = 0;
	TCNT2 = 0;
}


bool continue_play()
{
	return true;
}

/*
 * Construieste header-ul unui fisier WAVE
 */
void build_header(DWORD data_size){
	
	ST_DWORD(buffer1, 	FCC('R','I','F','F'));
	ST_DWORD(buffer1 + 4, 	504 + data_size);
	ST_DWORD(buffer1 + 8, 	FCC('W','A','V','E'));
	
	ST_DWORD(buffer1 + 12,	FCC('f','m','t',' '));
	ST_DWORD(buffer1 + 16,	16);
	ST_WORD(buffer1 + 20, 1);
	ST_WORD(buffer1 + 22, 1);
	ST_DWORD(buffer1 + 24, 16000);
	ST_DWORD(buffer1 + 28, 16000);
	ST_WORD(buffer1 + 32, 1);
	ST_WORD(buffer1 + 34, 8);
	
	ST_DWORD(buffer1 + 36, FCC('d','a','t','a'));
	ST_DWORD(buffer1 + 40, 468 + data_size);
	
	memset(buffer1 + 44, 0, 212);
}

/*
 * Incarca header-ul unui fisier WAVE
 *
 * @return DWORD
 * 	0 => format invalid
 * 	1 => eroare I/O
 * 	>1 => numarul de sample-uri
 */
DWORD load_header(void)
{
	DWORD size;
	WORD ret;

	// citeste header-ul (12 octeti)
	if(pf_read(BUF_FRONT, 12, &ret))
		return 1;

	if(ret != 12 || LD_DWORD(BUF_FRONT + 8) != FCC('W','A','V','E'))
		return 0;

	for(;;)
	{
		// citeste chunk ID si size
		pf_read(BUF_FRONT, 8, &ret);
		if(ret != 8)
			return 0;

		size = LD_DWORD(&BUF_FRONT[4]);


		// verifica FCC
		switch(LD_DWORD(&BUF_FRONT[0]))
		{
			// 'fmt ' chunk
			case FCC('f','m','t',' '):
				// verifica size

				if(size > 100 || size < 16) return 0;
				// citeste continutul
				pf_read(BUF_FRONT, size, &ret);
				// verifica codificarea
				if (ret != size) return 0;
				if (BUF_FRONT[0] != 1) {return 0;}

				// verifica numarul de canale
				if(BUF_FRONT[2] != 1 && BUF_FRONT[2] != 2) return 0;
				// verifica rezolutia
				if(BUF_FRONT[14] != 8 && BUF_FRONT[14] != 16) return 0;

				// seteaza sampling rate-ul
				OCR0A = (BYTE)(F_CPU / 8 / LD_WORD(&BUF_FRONT[4])) - 1;
				break;

			// 'data' chunk => incepe redarea
			case FCC('d','a','t','a'):
				return size;

			// 'LIST' chunk => skip
			case FCC('L','I','S','T'):
			// 'fact' chunk => skip
			case FCC('f','a','c','t'):
				pf_lseek(fs.fptr + size);
				break;

			// chunk necunoscut => eroare
			default:
				return 0;
		}
	}

	return 0;
}

/*
 * Functie care reda un fisier audio
 *
 * path - calea absoluta a fisierului
 *
 * @return UINT
 *	FR_OK daca a rulat cu succes fisierul
 */
UINT play(const char *path)
{
	FRESULT ret;

	if((ret = pf_open(path)) == FR_OK)
	{
		sei();
		WORD bytes_read;

		// incarca header-ul fisierului
		DWORD current_size = load_header();
		if(current_size < buf_size) {
			return FR_NO_FILE;
		}

		// align to sector boundary
		ret = pf_lseek((fs.fptr + 511) & ~511);
		if(ret != FR_OK) {
			return ret;
		}

		// fill front buffer
		ret = pf_read(BUF_FRONT, buf_size, &bytes_read);
		if(ret != FR_OK) {
			return ret;
		}
		if(bytes_read < buf_size)
			return ret;

		// reset front buffer index
		buf_pos = 0;

		// start output
		timer0_start();
		timer1_start();
		DDRD |= (1 << PD4);

		while(continue_play())
		{
			uint8_t old_buf_front = buf_front;
			
			// fill back buffer
			ret = pf_read(BUF_BACK, buf_size, &bytes_read);
			if(ret != FR_OK)
				break;
			if(bytes_read < buf_size)
				break;

			// wait for buffer swap
			while(old_buf_front == buf_front) ;
		}

		// stop output
		DDRD &= ~(1 << PD4);
		timer1_stop();
		timer0_stop();
		cli();	
	}

	return ret;
}

/*
 * Cauta si returneaza al n-lea fisier din folder-ul primit
 */
void get_music(int n, const char *folder, char *filename)
{
	DIR dir;
	FILINFO fil;

	char i = 0;

	pf_opendir(&dir, MUSIC);
	for(i = 0; i < n; i++) {
		pf_readdir(&dir, &fil);
		if(fil.fname[0] == 0x00) {
			filename[0] = 0x00;
			return;
		}
	}

	filename[0] = 0;
	strcat(filename, folder);
	strcat(filename, "/");
	strcat(filename, fil.fname);
}

void recordsomething()
{
	WORD written;
	unsigned char write2;
	DWORD data_size = 0;

	sprintf(filename, "music/sample0.wav");

	build_header(0);

	pf_open(filename);

	// scrie header in fisier
	pf_write(buffer1, 256, &written);
	memset(buffer1, 0, 256);
	pf_write(buffer1, 256, &written);

	converted = 0;
	sent = 0;
	
	// set OCR0
	OCR0A = F_CPU/8/16000 - 1;
	counter = 0;
	write1 = 1;
	
	sei();

	timer0_start();
	timer1_start();
	timer2_start();

	write2 = write1;
	
	do {
		while (write2 == write1);

		if (write2)
		pf_write(buffer1, WRITE_BUF_SIZE, &written);
		else
		pf_write(buffer2, WRITE_BUF_SIZE, &written);
		
		sent ++;
		data_size += WRITE_BUF_SIZE;
		
		write2 = write1;

		PORTD |= (1 << PD1);
		
	} while(data_size < 180000);

	PORTD &= ~(1 << PD1);

	timer0_stop();
	timer1_stop();
	timer2_stop();
	cli();

	ADCSRA &= ~(1<<ADEN);
	pf_write(buffer1, 0, &written);

	pf_open(filename);
	
	build_header(data_size);
	pf_write(buffer1, 256, &written);
	memset(buffer1, 0, 256);
	pf_write(buffer1, 256, &written);
	pf_write(buffer1, 0, &written);

}

int main() {

	IO_init();
	flag = false;

	PCICR |= (1 << PCIE2);
	PCMSK2 |= (1 << PCINT22);

	// secventa LED uri start program
        PORTD ^= (1 << PD0);
        _delay_ms(500);
        PORTD ^= (1 << PD0);
        PORTD ^= (1 << PD1);
        _delay_ms(500);
        PORTD ^= (1 << PD1);
        PORTD ^= (1 << PD2);
        _delay_ms(500);
        PORTD ^= (1 << PD2);

/*
	// COD DE TEST PENTRU SENZORUL DE SUNET:
	// cat timp detecteaza sunet, ledul de pe pd1 sta aprins, altfel sta stins

	while(1) {
		if ((PINC & (1 << PC6)) != 0) {
			PORTD &= ~(1 << PD1);
			recording = 0;
		}
		else {
			PORTD |= (1 << PD1);
			recording = 1;
			_delay_ms(1000);
		}		
	}*/


	while(pf_mount(&fs) != FR_OK) {
		// wait a while and retry
		_delay_ms(1000);
	}

	
	ADC_init();

	recordsomething();
	PORTD |= (1 << PD2);
	play("music/sample0.wav");
	PORTD &= ~(1 << PD2);

    	return 0;
}

